@@ -1,17 +1,17 @@
 import uuid
 from typing import Optional
-from contextvars import ContextVar
+from contextvars import ContextVar, Token
 
 ctx_session_id: ContextVar[str] = ContextVar('ck.session_id')
-ctx_session_timeout: ContextVar[int] = ContextVar('ck.session_timeout')
+ctx_session_timeout: ContextVar[float] = ContextVar('ck.session_timeout')
 
 
 class SessionContext:
-    def __init__(self, session: str, timeout: int):
+    def __init__(self, session: str, timeout: float):
         self.session = session
         self.timeout = timeout
-        self.token1 = None
-        self.token2 = None
+        self.token1: Optional[Token[str]] = None
+        self.token2: Optional[Token[float]] = None
 
     def __enter__(self) -> str:
         self.token1 = ctx_session_id.set(self.session)
